#include "Scenario.h"

#include <sstream>
#include <iomanip>

#include "srcMap/point.h"
#include "srcMap/carte.h"
#include "Vehicle.h"
#include "ElectricCircuit.h"
#include "ScenarioLoader.h"
#include "ScenarioData.h"
#include "VehicleSolverAbsoluteOptimal.h"
#include "VehicleSolverIgnoreStationOptimal.h"
#include "VehicleSolverWait.h"
#include "ObjectiveFunctionDuration.h"
#include "SystemSolverImmediate.h"
#include "SystemSolution.h"
#include "ISystemSolver.h"
#include "PerformanceAnalysis.h"
#include "SystemSolverCAStar.h"
#include "SystemSolverWHCAStar.h"
#include "ArrangementNone.h"
#include "ArrangementCascade.h"
#include "ArrangementPermutation.h"

const string Scenario::sSolverTypeAssociation[Scenario::eMAX_SOLVERTYPE] =
{"Init"
,"A*"
,"A* (includes waiting)"
,"HCA* (First in)"
,"HCA* (Cascade)"
,"HCA* (Permutation)"
,"WHCA* (small) (First in)"
,"WHCA* (medium) (First in)"
,"WHCA* (large) (First in)"
,"WHCA* (small) (Cascade)"
,"WHCA* (medium) (Cascade)"
,"WHCA* (large) (Cascade)"
,"WHCA* (small) (Permutation)"
,"WHCA* (medium) (Permutation)"
,"WHCA* (large) (Permutation)"
};

const string Scenario::sArrangementAssociation[Scenario::eMAX_ARRANGEMENT] =
{"(First in)"
,"(Cascade)"
,"(Permutation)"
};


Scenario::Scenario(IScenarioLoader* inScenarioLoader, Carte const & inRefCarte)
	:IScenario(inScenarioLoader)
{
	//Data
	mLoaderData = new ScenarioData();
	mRefScenarioLoader->Load(mLoaderData);

	//Objective function
	mObjectiveFunction.push_back(new ObjectiveFunctionDuration());

	//Arrangements possible
	mArrangement.push_back(new ArrangementNone()); //eNONE
	mArrangement.push_back(new ArrangementCascade()); //eCASCADE
	mArrangement.push_back(new ArrangementPermutation()); //ePERMUTATION


	//******SCENARIO CREATION*******
	//BASIC SCENARIOS
	////We use the first solver (that is fast) to initialise
	mSolverCircuit.push_back(this->GetRefData()->mElectricCircuit->clone());
	mVehicleSolver.push_back(new VehicleSolverAbsoluteOptimal(&inRefCarte, mSolverCircuit.back(), mObjectiveFunction.back()));
	mSystemSolver.push_back(new SystemSolverImmediate(mObjectiveFunction.back(), mVehicleSolver.back(), mSolverCircuit.back()));
	mSystemSolver.back()->GetCurrentSystemSolution()->mDescription = "A* (init abstract space for HA*)";

	//What people think they will get
	mSolverCircuit.push_back(this->GetRefData()->mElectricCircuit->clone());
	mVehicleSolver.push_back(new VehicleSolverAbsoluteOptimal(&inRefCarte, mSolverCircuit.back(), mObjectiveFunction.back()));
	mSystemSolver.push_back(new SystemSolverImmediate(mObjectiveFunction.back(), mVehicleSolver.back(), mSolverCircuit.back()));
	mSystemSolver.back()->GetCurrentSystemSolution()->mDescription = "Non-cooperation: HA* (no waiting at stations).";

	//What people will experience
	mSolverCircuit.push_back(this->GetRefData()->mElectricCircuit->clone());
	mVehicleSolver.push_back(new VehicleSolverWait(&inRefCarte, mObjectiveFunction.back(),mSolverCircuit.back()));
	mSystemSolver.push_back(new SystemSolverImmediate(mObjectiveFunction.back(), mVehicleSolver.back(), mSolverCircuit.back()));
	mSystemSolver.back()->GetCurrentSystemSolution()->mDescription = "Non-cooperation: HA* (including real waiting at stations).";
	
	
	//ORDERED SCENARIOS
	//HCA
	for (size_t i = 0; i < size_t(eMAX_ARRANGEMENT); i++)
	{
		//Waiting minimisation solver
		mSolverCircuit.push_back(this->GetRefData()->mElectricCircuit->clone());
		mVehicleSolver.push_back(new VehicleSolverIgnoreStationOptimal(&inRefCarte, mSolverCircuit.back(), mObjectiveFunction.back()));
		mSystemSolver.push_back(new SystemSolverCAStar(mObjectiveFunction.back(), mVehicleSolver.back(), mArrangement[i], mSolverCircuit.back()));
		mSystemSolver.back()->GetCurrentSystemSolution()->mDescription = "HCA* " + sArrangementAssociation[i];
	}

	//WHCA
	const size_t sizeWindow = 3;
	float fTimeWindow[sizeWindow] = { 500.0f ,1000.0f,2000.0f };
	string strTimeWindowAssociation[sizeWindow] = { "small","medium","large" };
	//Window will be divided by this factor for recomputing
	float stepFactor	= 2.0f;

	bool uqamMap = true;
	if (uqamMap)
	{
		fTimeWindow[0] = 25.0f;
		fTimeWindow[1] = 50.0f;
		fTimeWindow[2] = 100.0f;
	}
	for (size_t i = 0; i < size_t(eMAX_ARRANGEMENT); i++)
	{
		for (size_t j = 0;j < sizeWindow;j++)
		{
			//Sweep window (small)
			mSolverCircuit.push_back(this->GetRefData()->mElectricCircuit->clone());
			mVehicleSolver.push_back(new VehicleSolverIgnoreStationOptimal(&inRefCarte, mSolverCircuit.back(), mObjectiveFunction.back()));
			mSystemSolver.push_back(new SystemSolverWHCAStar(mObjectiveFunction.back(), mVehicleSolver.back(), mArrangement[i], mSolverCircuit.back(), fTimeWindow[j], fTimeWindow[j] / stepFactor));
			mSystemSolver.back()->GetCurrentSystemSolution()->mDescription = "WHCA* " + strTimeWindowAssociation[j] + " " + sArrangementAssociation[i];
		}
	}
}

Scenario::~Scenario()
{
	delete mLoaderData;
}

inline std::string Scenario::toStringPerformance() const
{
	std::ostringstream ss;
	ss  << std::setw(40) << left << "Performance for each" << " " << PerformanceAnalysis::toStringHeader() << endl;

	for (size_t i = 0; i < mSystemSolver.size(); i++)
	{
		ss << std::setw(40) << left << sSolverTypeAssociation[i] << " " << *mSystemSolver[i]->GetPerformanceAnalysis() << endl;
	}
	return ss.str();
}

